#include <cstdlib>
#include <cmath>

// ⏕Cu
#include "Gg.h"
using namespace gg;

// _f[^
static const int pointn = 20;                               // _̐
static const GLfloat pointSize = 5.0f;                      // _̑傫
static const GLfloat point0[] = { -0.9f,  0.0f,  0.0f };    // n_̈ʒu
static const GLfloat point1[] = {  0.9f,  0.0f,  0.0f };    // I_̈ʒu

// _f[^
static class PointBuffer
{
  // _obt@IuWFNg
  GgBuffer<GLfloat[4]> position;

  // f[^̎QƗpɎgeNX`obt@IuWFNg
  GgBuffer<GLfloat[4]> table;

  // eNX`obt@IuWFNgQƂeNX`
  GLuint tex;

public:
  
  // fXgN^
  ~PointBuffer(void) {}
  
  // RXgN^
  //    n: _, pos: _̈ʒu
  PointBuffer(unsigned int n)
  {
    // _obt@IuWFNg̃mۂ
    position.load(GL_ARRAY_BUFFER, n, 0, GL_DYNAMIC_COPY);

    // 쐬_obt@IuWFNgɏlݒ肷
    GLfloat (*p)[4] = (GLfloat (*)[4])glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
    // ŏ̓_
    p[0][0] = point0[0];
    p[0][1] = point0[1];
    p[0][2] = point0[2];
    p[0][3] = 0.0f;
    // Ԃ̓_
    for (unsigned int i = 1; i < n - 1; ++i)
    {
      GLfloat t = (GLfloat)i / (GLfloat)(pointn - 1);
      
      p[i][0] = point0[0] * (1.0f - t) + point1[0] * t;
      p[i][1] = point0[1] * (1.0f - t) + point1[1] * t;
      p[i][2] = point0[2] * (1.0f - t) + point1[2] * t;
      p[i][3] = 1.0f;
    }
    // Ō̓_
    p[n - 1][0] = point1[0];
    p[n - 1][1] = point1[1];
    p[n - 1][2] = point1[2];
    p[n - 1][3] = 0.0f;

    glUnmapBuffer(GL_ARRAY_BUFFER);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    
    // f[^̎QƗpɎgeNX`obt@IuWFNg̃mۂ
    table.load(GL_TEXTURE_BUFFER, n, 0, GL_DYNAMIC_COPY);

    // eNX`obt@IuWFNgQƂeNX`쐬
    glActiveTexture(GL_TEXTURE0);
    glGenTextures(1, &tex);
    glBindTexture(GL_TEXTURE_BUFFER, tex);
    glTexBuffer(GL_TEXTURE_BUFFER, GL_RGBA32F, table.buf());
    glBindTexture(GL_TEXTURE_BUFFER, 0);
  }

  // `
  void draw(GLint pvLoc, GLenum mode)
  {
    glBindBuffer(GL_ARRAY_BUFFER, position.buf());
    glEnableVertexAttribArray(pvLoc);
    glVertexAttribPointer(pvLoc, 4, GL_FLOAT, GL_FALSE, 0, 0);
    glDrawArrays(mode, 0, position.num());
    glDisableVertexAttribArray(pvLoc);
  }

  // obt@̓_ (x, y) ɋ߂̂T
  int pick(GLfloat x, GLfloat y, GLfloat dx, GLfloat dy) const
  {
    glBindBuffer(GL_ARRAY_BUFFER, position.buf());
    GLfloat (*p)[4] = (GLfloat (*)[4])glMapBuffer(GL_ARRAY_BUFFER, GL_READ_ONLY);
    for (unsigned int i = 0; i < position.num(); ++i)
    {
      if (fabs(p[i][0] - x) <= dx && fabs(p[i][1] - y) <= dy)
      {
        glUnmapBuffer(GL_ARRAY_BUFFER);
        return i;
      }
    }
    glUnmapBuffer(GL_ARRAY_BUFFER);
    return -1;
  }
  
  // obt@ i Ԗڂ̓_̈ʒu (x, y) ɐݒ肷
  void move(int i, GLfloat x, GLfloat y, GLfloat value) const
  {
    glBindBuffer(GL_ARRAY_BUFFER, position.buf());
    GLfloat p[] = { x, y, 0.0, value };
    glBufferSubData(GL_ARRAY_BUFFER, sizeof (GLfloat[4]) * i, sizeof p, p);
  }

  // vZ̎s
  void calculate(unsigned int loop = 1)
  {
    for (unsigned int i = 0; i < loop; ++i)
    {
      // _obt@IuWFNgƃeNX`obt@IuWFNg
      glBindBuffer(GL_ARRAY_BUFFER, position.buf());
      glBindBuffer(GL_TEXTURE_BUFFER, table.buf());

      // _obt@IuWFNgeNX`IuWFNgɃRs[
      glCopyBufferSubData(GL_ARRAY_BUFFER, GL_TEXTURE_BUFFER, 0, 0, sizeof (GLfloat[4]) * position.num());

      // _obt@IuWFNgƃeNX`obt@IuWFNg
      glBindBuffer(GL_TEXTURE_BUFFER, 0);
      glBindBuffer(GL_ARRAY_BUFFER, 0);

      // _obt@IuWFNg position ̃^[QbgƂĎw肷
      glBindBufferBase(GL_TRANSFORM_FEEDBACK_BUFFER, 0, position.buf());

      // Transform Feedback Jn
      glBeginTransformFeedback(GL_POINTS);

      // eNX`obt@IuWFNg̃eNX`
      glBindTexture(GL_TEXTURE_BUFFER, tex);

      // X^CU~߂ČvZs
      glEnable(GL_RASTERIZER_DISCARD);
      glDrawArrays(GL_POINTS, 0, position.num());
      glDisable(GL_RASTERIZER_DISCARD);

      // Transform Feedback I
      glEndTransformFeedback();

      // eNX`obt@IuWFNg̃eNX`
      glBindTexture(GL_TEXTURE_BUFFER, 0);
    }
  }

} *pointBuffer = 0;                           // _obt@IuWFNg

// VF[_
static class PointShader
  : public GgShader
{
  // attribute ϐ̏ꏊ
  GLint positionLoc;
  
  // uniform ϐ̏ꏊ
  GLint colorLoc;

public:

  // fXgN^
  ~PointShader(void) {}
  
  // RXgN^
  PointShader(
    const char *vert = "point.vert",
    const char *frag = "point.frag",
    const char *position = "position",
    const char *color = "color",
    const char *fragcolor = "FragColor"
    )
    : GgShader(vert, frag)
    , positionLoc(glGetAttribLocation(getProgram(), position))
    , colorLoc(glGetUniformLocation(getProgram(), color))
  {
    glBindFragDataLocation(getProgram(), 0, fragcolor);
  }
  
  // Fݒ肷
  void setColor(GLfloat r, GLfloat g, GLfloat b, GLfloat a = 1.0f) const
  {
    glUniform4f(colorLoc, r, g, b, a);
  }
  
  // attribute ϐ pos ̏ꏊ𓾂
  GLint getPositionLoc(void) const
  {
    return positionLoc;
  }
  
} *pointShader = 0;

static class ForceShader
  : public GgShader
{
  // uniform ϐ̏ꏊ
  GLint tableLoc;

public:

  // fXgN^
  ~ForceShader(void) {}
  
  // RXgN^
  ForceShader(
    const char *vert = "force.vert",
    const char *position = "position",
    const char *table = "table"
    )
  {
    const char *varyings[] = { position };

    load(vert, 0, 0, 0, 0, sizeof varyings / sizeof varyings[0], varyings);
    tableLoc = glGetUniformLocation(getProgram(), table);
  }

  // vOIuWFNggp
  void use(void) const
  {
      GgShader::use();
      glUniform1i(tableLoc, 0);
  }
    
} *forceShader = 0;

// n
static void cleanup(void)
{
  // _obt@IuWFNg폜
  delete pointBuffer;
  
  // vOIuWFNg폜
  delete pointShader;
  
  // ͂̌vZp̃VF[_폜
  delete forceShader;
}

// 
static void init(void)
{
  // Q[OtBbNX_̓sɂƂÂ
  ggInit();
  
  // _obt@IuWFNg쐬
  pointBuffer = new PointBuffer(pointn);
  
  // vOIuWFNgiVF[_j쐬
  pointShader = new PointShader();
  
  // ͂̌vZp̃VF[_쐬
  forceShader = new ForceShader();

  // OpenGL ̏ݒ
  glClearColor(1.0, 1.0, 1.0, 1.0);
  
  // n
  atexit(cleanup);
}

// Aj[Vۂ
static bool animate = false;

// ʂɐ}``悷
static void display(void)
{
  // ʃNA
  glClear(GL_COLOR_BUFFER_BIT);
  
  // _̕`
  pointShader->use();
  pointShader->setColor(0.0f, 0.0f, 0.0f);
  pointBuffer->draw(pointShader->getPositionLoc(), GL_LINE_STRIP);
  pointShader->setColor(1.0f, 0.0f, 0.0f);
  glPointSize(pointSize);
  pointBuffer->draw(pointShader->getPositionLoc(), GL_POINTS);

  // vZ̎s
  if (animate)
  {
    forceShader->use();
    pointBuffer->calculate(10);
  }

  // _uobt@O
  glutSwapBuffers();
}

// EBhE̒Sʒu
static GLfloat cx, cy;

// EBhẼI[vETCY̏
static void resize(int w, int h)
{
  // EBhEŜr[|[gi\̈jɂ
  glViewport(0, 0, w, h);
  
  // EBhE̕ƍoĂ
  cx = w * 0.5f;
  cy = h * 0.5f;
}

// ꂽ}EX{^
static int pressed;

// hbOĂ_
static int hit = -1;

// 邱ƂȂȂ̏
static void idle(void)
{
  // ʂ̍ĕ`s
  glutPostRedisplay();
}

// }EX̃{^Ƃ̏
static void mouse(int button, int state, int x, int y)
{
  // ꂽ{^oĂ
  pressed = button;
  
  switch (button)
  {
  case GLUT_LEFT_BUTTON:
    if (state == GLUT_DOWN)
    {
      // {^Ƃ̏
      hit = pointBuffer->pick((GLfloat)x / cx - 1.0f, 1.0f - (GLfloat)y / cy, pointSize / cx, pointSize / cy);
      if (!animate && hit >= 0) glutIdleFunc(idle);
    }
    else
    {
      // {^𗣂Ƃ̏
      pointBuffer->move(hit, (GLfloat)x / cx - 1.0f, 1.0f - (GLfloat)y / cy, 1.0);
      hit = -1;
      if (!animate) glutIdleFunc(0);
    }
    break;
  case GLUT_MIDDLE_BUTTON:
    if (state == GLUT_DOWN)
    {
      // {^Ƃ̏
    }
    else
    {
      // {^𗣂Ƃ̏
    }
    break;
  case GLUT_RIGHT_BUTTON:
    if (state == GLUT_DOWN)
    {
      // E{^Ƃ̏
    }
    else
    {
      // E{^𗣂Ƃ̏
    }
    break;
  default:
    break;
  }
}

// }EX̃hbȌ
static void motion(int x, int y)
{
  switch (pressed)
  {
  case GLUT_LEFT_BUTTON:
    // {^ŃhbȌ
    if (hit >= 0)
    {
      pointBuffer->move(hit, (GLfloat)x / cx - 1.0f, 1.0f - (GLfloat)y / cy, 0.0);
    }
    break;
  case GLUT_MIDDLE_BUTTON:
    // {^ŃhbȌ
    break;
  case GLUT_RIGHT_BUTTON:
    // E{^ŃhbȌ
    break;
  default:
    break;
  }
}

// L[{[h^CvƂ̏
static void keyboard(unsigned char key, int x, int y)
{
  switch (key) {
  case ' ':
    animate = !animate;
    if (animate)
      glutIdleFunc(idle);
    else
      glutIdleFunc(0);
    break;
  case '\033':
  case 'q':
  case 'Q':
    // ESC L[, q, Q ^CvI
    exit(0);
  default:
    break;
  }
}

// CvO
int main(int argc, char *argv[])
{
  glutInitWindowSize(500, 500);
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
  glutCreateWindow("physics");
  glutDisplayFunc(display);
  glutReshapeFunc(resize);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keyboard);
  init();
  glutMainLoop();

  return 0;
}
